import os

import click
import torch
import torch.nn as nn
import torch.optim as optim
import torch.nn.functional as F
import torchvision
import torchvision.transforms as transforms
import torch.distributed as dist
from torch.utils.tensorboard.writer import SummaryWriter


def evaluate(model, test_loader, device):
    with torch.no_grad():
        model.eval()

        correct = total = 0
        for x, y in test_loader:
            x = x.to(device)
            y = y.to(device)

            pred = model(x)

            max_val, max_args = torch.max(pred, 1)

            correct += torch.sum(max_args == y).item()
            total += len(y)

    accuracy = correct / total

    return accuracy


@click.command()
@click.option("--data_dir", required=True)
@click.option("--output_dir", required=True)
def main(data_dir, output_dir):
    local_rank = int(os.environ["LOCAL_RANK"])

    device = (
        torch.device("cuda", local_rank)
        if torch.cuda.is_available()
        else torch.device("cpu", local_rank)
    )
    backend = "nccl" if torch.cuda.is_available() else "gloo"
    dist.init_process_group(backend)

    log_dir = os.path.join(output_dir, "log")
    model_path = os.path.join(output_dir, "model.pt")

    summary_writer = SummaryWriter(log_dir=log_dir)

    model = torchvision.models.mobilenet_v3_small(num_classes=100)

    criterion = nn.CrossEntropyLoss()
    optimizer = optim.Adam(model.parameters(), lr=0.01)
    transform = transforms.Compose(
        [
            transforms.ToTensor(),
            transforms.Normalize(mean=[0.485, 0.456, 0.406], std=[0.229, 0.224, 0.225]),
        ]
    )

    trainset = torchvision.datasets.CIFAR100(
        root=data_dir, train=True, download=False, transform=transform
    )
    train_sampler = torch.utils.data.distributed.DistributedSampler(trainset)
    trainloader = torch.utils.data.DataLoader(
        trainset, batch_size=100, shuffle=False, num_workers=2, sampler=train_sampler
    )

    testset = torchvision.datasets.CIFAR100(
        root=data_dir, train=False, download=False, transform=transform
    )
    testloader = torch.utils.data.DataLoader(
        testset, batch_size=100, shuffle=False, num_workers=2
    )

    model = model.to(device)
    model = torch.nn.parallel.DistributedDataParallel(model)

    print("Start Training")
    for epoch in range(10):
        model.train()

        train_sampler.set_epoch(epoch)
        running_loss = 0.0
        for i, data in enumerate(trainloader, 0):
            inputs, labels = data[0].to(device), data[1].to(device)

            outputs = model(inputs)
            loss = criterion(outputs, labels)

            optimizer.zero_grad()
            loss.backward()
            optimizer.step()

            running_loss += loss.item()

        running_loss /= len(trainloader)
        running_loss = torch.tensor(running_loss).to(device)
        avg_loss = running_loss.clone()
        dist.all_reduce(avg_loss, op=dist.ReduceOp.SUM)
        avg_loss /= dist.get_world_size()

        if local_rank == 0:
            summary_writer.add_scalar("loss", avg_loss.item(), global_step=epoch + 1)
            print("Epoch {} Loss: {:.4f}".format(epoch + 1, avg_loss.item()))

        accuracy = evaluate(model, testloader, device)

        if local_rank == 0:
            summary_writer.add_scalar("accuracy", accuracy, global_step=epoch + 1)
            print("Epoch {} Accuracy: {:.4f}".format(epoch + 1, accuracy))

    print("Finished Training")

    if local_rank == 0:
        torch.save(model.state_dict(), model_path)
        print("Saved model to {}".format(model_path))

    dist.destroy_process_group()


if __name__ == "__main__":
    main()